#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fstream>

#include <string>
#include "CAM.h"

/* camera constants from A.fin */
#define CCDSIZE_X 6.3840
#define CCDSIZE_Y 4.8048
#define TOTAL_PIXEL_X  640
#define TOTAL_PIXEL_Y  480

#define f_DEFAULT 1.1405936e+01
#define Asp_DEFAULT 1.0022303e+00
#define Cx_DEFAULT 3.4342111e+02
#define Cy_DEFAULT 2.8150982e+02

using namespace std;

CAM_Finland::CAM_Finland()
{
  // default constructor
  NDX=TOTAL_PIXEL_X; NDY=TOTAL_PIXEL_Y;
  Sx=CCDSIZE_X; Sy=CCDSIZE_Y;
  dpx=Sx/NDX; dpy=Sy/NDY;
  Cx=Cx_DEFAULT; Cy=Cy_DEFAULT; Asp=Asp_DEFAULT;f=f_DEFAULT;
  
  Tx=Ty=Tz=0;
  Rx=Ry=Rz=0;

  for (int i=0;i<3;i++)
    for (int j=0;j<3;j++)
      rotm[i][j]=0;
  rotm[0][0]=rotm[1][1]=rotm[2][2]=1.0;
  
  // Initialize projection matrix
  updateProjMatrixAndCamCenter();
}

/* This function reads the camera parameters from an output file of
   Janne Heikkila's camera calibration algorithm */

CAM_Finland::CAM_Finland(CAM_Finland &cam)
{
  //name
  sprintf(name,"NEWCAM");
  
  //constants
  NDX=cam.NDX;
  NDY=cam.NDY;
  Sx=cam.Sx;
  Sy=cam.Sy;
  dpx=cam.dpx;
  dpy=cam.dpy;
  //intrinsics
  Asp=cam.Asp;
  f=cam.f;
  Cx=cam.Cx;
  Cy=cam.Cy;
  Rad1=cam.Rad1;
  Rad2=cam.Rad2;
  Tan1=cam.Tan1;
  Tan2=cam.Tan2;

  a1=cam.a1;
  a2=cam.a2;
  a3=cam.a3;
  a4=cam.a4;
  a5=cam.a5;
  a6=cam.a6;
  a7=cam.a7;
  a8=cam.a8;

  //extrinsics
  Tx=cam.Tx;
  Ty=cam.Ty;
  Tz=cam.Tz;
  Rx=cam.Rx;
  Ry=cam.Ry;
  Rz=cam.Rz;

  //other non-independent parameters
  Rquat=cam.Rquat;
  for (int i=0;i<3;i++)
    for (int j=0;j<3;j++)
      rotm[i][j]=cam.rotm[i][j];
  P=cam.P;
  invP=cam.invP;
  C=cam.C;
}

CAM_Finland::CAM_Finland(char *fname)
{
  
  ifstream infile(fname);

  if (!infile.good())
    error("Can't open input camera file", fname);

  strcpy(name,fname);
  
  //constants
  infile >> NDX;
  infile >> NDY;
  infile >> Sx;
  infile >> Sy;
  
  dpx=Sx/NDX;
  dpy=Sy/NDY;

  // intrinsic parameters
  
  infile >> Asp; //NOTE: Asp is the Aspect ratio here, not Sx!!!
  infile >> f;  // effective focal length
  infile >> Cx; //principal point
  infile >> Cy;
  // lens distortion parameters
  infile >> Rad1;
  infile >> Rad2;
  infile >> Tan1;
  infile >> Tan2;
  // inverse lens distortion parameters
  infile >> a1;
  infile >> a2;
  infile >> a3;
  infile >> a4;
  infile >> a5;
  infile >> a6;
  infile >> a7;
  infile >> a8;

  // extrinsic parameters
  
  infile >> Tx;
  infile >> Ty;
  infile >> Tz;
  infile >> Rx;
  infile >> Ry;
  infile >> Rz;

  std::string comment;
  infile >> comment;

  double q1,q2,q3,q4;
  bool use_quat=false;
  if (comment=="#USE_QUATERNION") {
    //use quaternion, the Euler angles Rx,Ry and Rz are ignored
    infile>>q1;
    infile>>q2;
    infile>>q3;
    infile>>q4;
    //quaternion
    Rquat=Quaternion(q1,q2,q3,q4);
    // equivalent rotation matrix
    Rquat.getRowMatrix(rotm);
    use_quat=true;

    //update Rx,Ry,Rz
    rotMatrix2Euler(rotm,Rx,Ry,Rz);
  } else {
    //NOTE: input Rx,Ry,Rz are in degrees rather than radians!!!
    Rx=Rx*M_PI/180.0;
    Ry=Ry*M_PI/180.0;
    Rz=Rz*M_PI/180.0;
  
    double sa = sin(Rx);
    double ca = cos(Rx);
    double sb = sin(Ry);
    double cb = cos(Ry);
    double sg = sin(Rz);
    double cg = cos(Rz);

    r1 = cb * cg;
    r2 = cg * sa * sb - ca * sg;
    r3 = sa * sg + ca * cg * sb;
    r4 = cb * sg;
    r5 = sa * sb * sg + ca * cg;
    r6 = ca * sb * sg - cg * sa;
    r7 = -sb;
    r8 = cb * sa;
    r9 = ca * cb;
   
    //update quaternion
    Rquat=QuatFromRowMatrix(rotm);
    

  } 

  
  // Initialize projection matrix
  updateProjMatrixAndCamCenter();

  if (use_quat)
    printf("*** Camera File %s read, use Quaternion for orientation ***\n",fname);
  else
    printf("*** Camera File %s read, use Euler angles for orientation ***\n",fname);
}

void CAM_Finland::writecamfile(char *fname,bool use_quaternion)
{
  ofstream outfile(fname);

  if (!outfile.good())
    error("Can't open output camera file", fname);

  //constants
  outfile << NDX << endl;
  outfile << NDY << endl;
  outfile << Sx << endl;
  outfile << Sy << endl;
  
  // intrinsic parameters
  
  outfile << Asp << endl; //NOTE: Asp is the Aspect ratio here, not Sx!!!
  outfile << f << endl;  // effective focal length
  outfile << Cx << endl; //principal point
  outfile << Cy << endl;
  // lens distortion parameters
  outfile << Rad1 << endl;
  outfile << Rad2 << endl;
  outfile << Tan1 << endl;
  outfile << Tan2 << endl;
  // inverse lens distortion parameters
  outfile << a1 << endl;
  outfile << a2 << endl;
  outfile << a3 << endl;
  outfile << a4 << endl;
  outfile << a5 << endl;
  outfile << a6 << endl;
  outfile << a7 << endl;
  outfile << a8 << endl;
  
  // extrinsic parameters
  outfile << Tx << endl;
  outfile << Ty << endl;
  outfile << Tz << endl;

  // output Rx,Ry Rz in degrees
  double Rx_deg = Rx/M_PI * 180.0;
  double Ry_deg = Ry/M_PI * 180.0;
  double Rz_deg = Rz/M_PI * 180.0;
  
  outfile << Rx_deg << endl;
  outfile << Ry_deg << endl;
  outfile << Rz_deg << endl;

  if (use_quaternion) {
    //output quaternion
    outfile<<"#USE_QUATERNION"<<endl;
    outfile<<Rquat[0]<<endl;
    outfile<<Rquat[1]<<endl;
    outfile<<Rquat[2]<<endl;
    outfile<<Rquat[3]<<endl;
  }    
} 


/* This routine converts from distorted to undistorted sensor plane coordinates */
void  CAM_Finland::
distorted_to_undistorted_sensor_coord (double Xd, double Yd, 
				       double &Xu, double &Yu)
{
  double rsq =Xd*Xd + Yd*Yd;
  double G = a5*rsq*rsq + a6*Xd*rsq + a7*Yd*rsq + a8*rsq + 1;
  Xu = (Xd+Xd*(a1*rsq+a2*rsq*rsq)+2.0*a3*Xd*Yd+a4*(rsq+2.0*Xd*Xd))/G;
  Yu = (Yd+Yd*(a1*rsq+a2*rsq*rsq)+2.0*a4*Xd*Yd+a3*(rsq+2.0*Yd*Yd))/G;
}



/************************************************************************/
/*
  This routine converts from undistorted to distorted sensor coordinates.
*/

void CAM_Finland::
undistorted_to_distorted_sensor_coord  (double Xu, double Yu, 
					double &Xd, double &Yd) 
{

  double rsq=Xu*Xu+Yu*Yu;
  double delta=Rad1*rsq+Rad2*rsq*rsq;

  Xd=Xu*(1+delta)+ 2.0*Tan1*Xu*Yu + Tan2*(rsq + 2*Xu*Xu);
  Yd=Yu*(1+delta)+ Tan1*(rsq + 2*Yu*Yu) + 2*Tan2*Xu*Yu;

}
