#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fstream.h>

#include "CAM.h"

/* This function reads the camera parameters from an output file of
   Tsai/Wilson's camera calibration algorithm */


CAM_Tsai::CAM_Tsai(char *fname)
{
  
  ifstream infile(fname);

  if (!infile.good())
    error("Can't open camera file", fname);

  strcpy(name,fname); 
  
  infile >> Ncx;
  infile >> Nfx;
  infile >> dx;
  infile >> dy;
  infile >> dpx;
  infile >> dpy;
  infile >> Cx;
  infile >> Cy;
  infile >> Asp;
  Sx=dpx*Nfx;
  Sy=dy * Nfx * Asp;
  printf("CAM_Tsai() Warning: Check actual CCD size Sx=%lf,Sy=%lf\n",Sx,Sy);
  
  infile >> f;
  infile >> kappa1;
  infile >> Tx;
  infile >> Ty;
  infile >> Tz;
  infile >> Rx;
  infile >> Ry;
  infile >> Rz;

  double sa = sin(Rx);
  double ca = cos(Rx);
  double sb = sin(Ry);
  double cb = cos(Ry);
  double sg = sin(Rz);
  double cg = cos(Rz);

  
  r1 = cb * cg;
  r2 = cg * sa * sb - ca * sg;
  r3 = sa * sg + ca * cg * sb;
  r4 = cb * sg;
  r5 = sa * sb * sg + ca * cg;
  r6 = ca * sb * sg - cg * sa;
  r7 = -sb;
  r8 = cb * sa;
  r9 = ca * cb;

 
  //update quaternion
  Rquat=QuatFromRowMatrix(rotm);

  // Initialize projection matrix
  updateProjMatrixAndCamCenter();
  
  printf("*** Camera File %s read ***\n",fname);
}

void CAM_Tsai::writecamfile(char *fname,bool)
{
  ofstream outfile(fname);

  if (!outfile.good())
    error("Can't open output camera file", fname);

  outfile << Ncx << endl;
  outfile << Nfx << endl;
  outfile << dx << endl;
  outfile << dy << endl;
  outfile << dpx << endl;
  outfile << dpy << endl;
  outfile << Cx << endl;
  outfile << Cy << endl;
  outfile << Asp << endl;
 
  outfile << f << endl;
  outfile << kappa1 << endl;
  outfile << Tx << endl;
  outfile << Ty << endl;
  outfile << Tz << endl;
  outfile << Rx << endl;
  outfile << Ry << endl;
  outfile << Rz << endl;

  double dummy = 0.0;
  outfile << dummy <<endl;
  outfile << dummy <<endl;
}


/* This routine converts from distorted to undistorted sensor plane coordinates */
void  CAM_Tsai::
distorted_to_undistorted_sensor_coord (double Xd, double Yd, 
				       double &Xu, double &Yu)
{
  double    distortion_factor;

  distortion_factor = 1 + kappa1 * (Xd*Xd + Yd*Yd);
  Xu = Xd * distortion_factor;
  Yu = Yd * distortion_factor;

}

/************************************************************************/
/* This cube root routine handles negative arguments (unlike cbrt).     */

static double  CBRT (double x)
{
  if (x == 0)
    return (0);
  else if (x > 0)
    return (pow (x, (double) 1.0 / 3.0));
  else
    return (-pow (-x, (double) 1.0 / 3.0));
}

/************************************************************************/
/*
  This routine converts from undistorted to distorted sensor coordinates.
  The technique involves algebraically solving the cubic polynomial
  
  Ru = Rd * (1 + kappa1 * Rd**2)

  using the Cardan method.
       
  Note: for kappa1 < 0 the distorted sensor plane extends out to a maximum
  barrel distortion radius of  Rd = sqrt (-1/(3 * kappa1)).
  
  To see the logic used in this routine try graphing the above polynomial
  for positive and negative kappa1's
*/

void CAM_Tsai::
undistorted_to_distorted_sensor_coord (double Xu, double Yu, 
				       double &Xd, double &Yd)
{
#define SQRT3   1.732050807568877293527446341505872366943
#define SQRT(x) sqrt(fabs(x))

  double    Ru, Rd, lambda, c, d, Q, R, D, S, T, sinT, cosT;

  if (((Xu == 0) && (Yu == 0)) || (kappa1 == 0)) 
    {
      Xd = Xu;
      Yd = Yu;
      return;
    }

  Ru = hypot (Xu, Yu);        /* SQRT(Xu*Xu+Yu*Yu) */

  c = 1 / kappa1;
  d = -c * Ru;

  Q = c / 3;
  R = -d / 2;
  D = Q*Q*Q + (R*R);

  if (D >= 0) 
    {               /* one real root */
      D = SQRT (D);
      S = CBRT (R + D);
      T = CBRT (R - D);
      Rd = S + T;

      if (Rd < 0) 
	{
	  Rd = SQRT (-1 / (3 * kappa1));
	  fprintf(stderr, "\nWarning: undistorted image point to distorted image point mapping limited by\n");
	  fprintf (stderr, "         maximum barrel distortion radius of %lf\n", Rd);
	  fprintf (stderr, "         (Xu = %lf, Yu = %lf) -> (Xd = %lf, Yd = %lf)\n\n",
		   Xu, Yu, Xu * Rd / Ru, Yu * Rd / Ru);
	}
    } 
  else 
    {                    /* three real roots */
      D = SQRT (-D);
      S = CBRT (hypot (R, D));
      T = atan2 (D, R) / 3;
      sinT=sin(T);
      cosT=cos(T);

      /* the larger positive root is    2*S*cos(T)                   */
      /* the smaller positive root is   -S*cos(T) + SQRT(3)*S*sin(T) */
      /* the negative root is           -S*cos(T) - SQRT(3)*S*sin(T) */

      Rd = -S * cosT + SQRT3 * S * sinT;      /* use the smaller positive root */
    }

  lambda = Rd / Ru;

  Xd = Xu * lambda;
  Yd = Yu * lambda;
}


